#!/bin/sh

# Copyright (c) 2014-2015 Oracle and/or its affiliates. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it would be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write the Free Software Foundation,
# Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# Author: Alexey Kodanev <alexey.kodanev@oracle.com>
#
# Virtual eXtensible Local Area Network (VXLAN) provides L2 networks
# over existed L3 networks. It is using UDP (port 8472) to encapsulate
# data packets. More information:
# http://tools.ietf.org/html/draft-mahalingam-dutt-dcops-vxlan-08
#
# Warning: Test assumes that machines don't have any existed VXLANs.
#          If machine has VXLANs, the test might fail or eventually delete
#          them in cleanup function. See "start_vni" variable which can
#          solve it.

ip_local=$(tst_ipaddr)
ip_vxlan_local="192.168.124.1"
ip6_vxlan_local="fe80::381c:c0ff:fea8:7c01"
mac_vxlan_local="3A:1C:C0:A8:7C:01"

ip_remote=$(tst_ipaddr rhost)
ip_vxlan_remote="192.168.124.2"
ip6_vxlan_remote="fe80::381c:c0ff:fea8:7c02"
mac_vxlan_remote="3A:1C:C0:A8:7C:02"

vxlan_max=5000
start_vni=16700000
clients_num=2
client_requests=500000
max_requests=3
net_load="TFO"

# In average cases (with small packets less then 150 bytes) vxlan slower
# by 10-30%. If hosts are too close to each one, e.g. connected to the same
# switch the performance can be slower by 50%. Set 60% as default, the above
# will be an error in VXLAN.
vxlan_threshold=60

# Destination address, can be unicast or multicast address
vxlan_dst_addr="uni"

while getopts :hsx:i:r:c:R:p:n:l:t:d:6 opt; do
	case "$opt" in
	h)
		echo "Usage:"
		echo "h        help"
		echo "s        use ssh to run remote cmds"
		echo "x n      n is a number of VXLANs for tc1 and tc2"
		echo "i n      start VNI to use"
		echo "r n      client requests for TCP performance test"
		echo "c n      clients run concurrently in TCP perf"
		echo "R n      num of reqs, after which conn.closed in TCP perf"
		echo "p x      x and x + 1 are ports in TCP perf"
		echo "n x      VXLAN network 192.168.x"
		echo "l x      network load: x is PING or TFO(tcp_fastopen)"
		echo "t x      VXLAN performance threshold, default is 60%"
		echo "d x      VXLAN destination address, 'uni' or 'multi'"
		echo "6        run over IPv6"
		exit 0
	;;
	s) TST_USE_SSH=1 ;;
	x) vxlan_max=$OPTARG ;;
	i) start_vni=$OPTARG ;;
	c) clients_num=$OPTARG ;;
	r) client_requests=$OPTARG ;;
	R) max_requests=$OPTARG ;;
	p) srv_port=$OPTARG ;;
	n)
		ip_vxlan_local="192.168.${OPTARG}.1"
		ip_vxlan_remote="192.168.${OPTARG}.2"
	;;
	l) net_load=$OPTARG ;;
	t) vxlan_threshold=$OPTARG ;;
	d) vxlan_dst_addr=$OPTARG ;;
	6) # skip, test_net library already processed it
	;;
	*)
		tst_brkm TBROK "unknown option: $opt"
	;;
	esac
done

cleanup_vxlans()
{
	tst_resm TINFO "cleanup vxlans..."
	local vxlans=`ip link | sed -nE 's/^[0-9]+: (ltp_vxl[0-9]+):.+/\1/p'`
	for vx in $vxlans; do
		ip link delete $vx
	done
}

TST_CLEANUP="cleanup_vxlans"
trap "tst_brkm TBROK 'test interrupted'" INT

vxlan_setup_subnet_uni()
{
	tst_kvercmp 3 10 0 && \
		tst_brkm TCONF "test must be run with kernel 3.10 or newer"

	[ "$(ip li add type vxlan help 2>&1 | grep remote)" ] || \
		tst_brkm TCONF "iproute doesn't support remote unicast address"

	local opt="remote $(tst_ipaddr rhost)"

	tst_resm TINFO "setup VxLANv$ipver with unicast address: '$opt'"

	ROD_SILENT "ip li add ltp_vxl0 type vxlan id $1 $opt"
	ROD_SILENT "ip li set ltp_vxl0 address $mac_vxlan_local"
	ROD_SILENT "ip addr add ${ip_vxlan_local}/24 dev ltp_vxl0"
	ROD_SILENT "ip li set up ltp_vxl0"

	opt="remote $(tst_ipaddr)"

	tst_rhost_run -s -c "ip li add ltp_vxl0 type vxlan id $2 $opt"
	tst_rhost_run -s -c "ip li set ltp_vxl0 address $mac_vxlan_remote"
	tst_rhost_run -s -c "ip addr add ${ip_vxlan_remote}/24 dev ltp_vxl0"
	tst_rhost_run -s -c "ip li set up ltp_vxl0"
}

vxlan_setup_subnet_multi()
{
	tst_check_cmds "od"
	local b1=$(($(od -An -d -N1 /dev/urandom) % 254 + 1))
	local b2=$(($(od -An -d -N1 /dev/urandom) % 254 + 1))
	local b3=$(($(od -An -d -N1 /dev/urandom) % 254 + 1))

	local opt=
	if [ "$TST_IPV6" ]; then
		opt="group ff05::$(printf '%x:%x%x' $b1 $b2 $b3)"
	else
		opt="group 239.$b1.$b2.$b3"
	fi

	tst_resm TINFO "setup VxLANv$ipver with multicast address: '$opt'"

	ROD_SILENT "ip li add ltp_vxl0 type vxlan id $1 $opt dev $(tst_iface)"
	ROD_SILENT "ip li set ltp_vxl0 address $mac_vxlan_local"
	ROD_SILENT "ip addr add ${ip_vxlan_local}/24 dev ltp_vxl0"
	ROD_SILENT "ip li set up ltp_vxl0"

	tst_rhost_run -s -c "ip li add ltp_vxl0 type vxlan id $2 $opt \
	                     dev $(tst_iface rhost)"
	tst_rhost_run -s -c "ip li set ltp_vxl0 address $mac_vxlan_remote"
	tst_rhost_run -s -c "ip addr add ${ip_vxlan_remote}/24 dev ltp_vxl0"
	tst_rhost_run -s -c "ip li set up ltp_vxl0"
}

vxlan_compare_netperf()
{
	local ret=0
	tst_netload $ip_vxlan_remote res_ipv4 $net_load || ret=1
	tst_netload ${ip6_vxlan_remote}%ltp_vxl0 res_ipv6 $net_load || ret=1

	ROD_SILENT "ip link delete ltp_vxl0"
	tst_rhost_run -s -c "ip link delete ltp_vxl0"
	[ "$ret" -eq 1 ] && return 1
	local vt="$(cat res_ipv4)"
	local vt6="$(cat res_ipv6)"

	tst_netload $ip_remote res_ipv4 $net_load || return 1

	local lt="$(cat res_ipv4)"
	tst_resm TINFO "time lan($lt) vxlan IPv4($vt) and IPv6($vt6) ms"

	per=$(( $vt * 100 / $lt - 100 ))
	per6=$(( $vt6 * 100 / $lt - 100 ))

	tst_resm TINFO "IPv4 VxLAN over IPv$ipver slower by $per %"
	tst_resm TINFO "IPv6 VxLAN over IPv$ipver slower by $per6 %"

	[ "$per" -ge "$vxlan_threshold" -o "$per6" -ge "$vxlan_threshold" ] \
		&& ret=1

	return $ret
}

tst_require_root

tst_kvercmp 3 8 0 && \
	tst_brkm TCONF "test must be run with kernel 3.8 or newer"

if [ "$TST_IPV6" ]; then
	tst_kvercmp 3 12 0 && \
		tst_brkm TCONF "test must be run with kernel 3.12 or newer"
fi

ipver=${TST_IPV6:-'4'}

tst_check_cmds "ip"

ip link add ltp_vxl type vxlan id $start_vni > /dev/null 2>&1
if [ $? -ne 0 ]; then
	tst_brkm TCONF "iproute2 or kernel doesn't support vxlan"
fi
ROD_SILENT "ip link delete ltp_vxl"
